﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Threading;
using System.Threading;
using System.Net;
using System.Net.Sockets;

namespace MyTcp
{
   public class Agent<T> where T : IShowable 
   {
      public ManualResetEvent _waitAck = new ManualResetEvent(true);
      
      public enum RequestStatus
      {
         WaitForLoginSignal = 1 ,
         WaitForAccountPassword = 2 ,
         ListenForRequest = 3 ,
         WaitForRegistration = 4,
         ListenAddFriend = 5,
         WaitForStatusChangeSignal = 6,
         WaitForAddFriendSignal = 7,
         NotifyNewFriend = 8,
      }           
      
      public Socket AgentSocket { get {return _socket ;} set{ _socket = value ;}}
      private Socket _socket;
      public byte[] _dataBuffer = new byte[1024];
      public byte[] _responseBuffer ;
      private RequestStatus _status ;
      public RequestStatus Status { get {return _status;} set{_status = value ;} }
      public T _windowRef;
      public bool _active ;  
      public int _tick = 0 ;
      public bool _waiting = false ;
      private int _id = 0 ;
      public int AgentID { get { return _id;} set { _id = value ;} }

      public static List<Agent<T>> agentList = new List<Agent<T>>();
      public List<Agent<T>> _listRef;
      
      public Agent()
      {
         _socket = null ; 
         _status = RequestStatus.WaitForLoginSignal ; 
         _windowRef = default(T) ;
         _active = true ;
      }
      
      public Agent(T w) 
      { 
         _socket = null ; 
         _status = RequestStatus.WaitForLoginSignal ; 
         _windowRef = w ;
         _active = true ;
      }
      
      public void InitAgent(T w, ref List<Agent<T>> l)
      {
         _windowRef = w ;
         _listRef = l ;
      }            

      protected void Send()
      {
         this.AgentSocket.Send(_responseBuffer) ;         
         // clear _responseBuffer after the data is sent
         if (_responseBuffer != null)
            Array.Clear(_responseBuffer, 0, _responseBuffer.Length); 
      }

      public void SendWait( byte[] bytes )
      {
         // Clear the response buffer first
         if (_responseBuffer != null)
            Array.Clear(_responseBuffer, 0, _responseBuffer.Length);         
         _responseBuffer = new byte[bytes.Length] ;
         Buffer.BlockCopy(bytes, 0, _responseBuffer, 0, bytes.Length) ;
         Send(); 
         Wait();
      }      
      
      public void Send( byte[] bytes )
      {
         // Clear the response buffer first
         if (_responseBuffer != null)
            Array.Clear(_responseBuffer, 0, _responseBuffer.Length);         
         _responseBuffer = new byte[bytes.Length] ;
         Buffer.BlockCopy(bytes, 0, _responseBuffer, 0, bytes.Length) ;
         Send();
         //Wait();         
      }   
      
      public void Wait()
      {  
         _waitAck.Reset() ;
         _waitAck.WaitOne() ;
         /*
         _tick = 0 ;
         TimerCallback tcb = new TimerCallback(Inc) ;
         Timer clock = new Timer(tcb, null, 0, 800) ;
         while(true)
            if( _tick > 1 ) {
               _tick = 0 ;
               break ;
            }*/
      }
            
      public void Inc(object o)
      { 
         ++_tick ;
      }                           
      
      public void ShowMessage(string s)
      {         
         _windowRef.ShowMessage(s) ;         
      }

      private UInt16 FindVacantPort()
      {
         List<int> usedPorts = new List<int>();
         Random r = new Random();
         int newPort = -1;

         TcpListener newListener;

         while (true)
         {
            newPort = (int)r.Next(1025, 65535); // be nice, don't use ports bellow 1025
            if (usedPorts.Contains(newPort))
               continue;
            try
            {
               newListener = new TcpListener(IPAddress.Any, newPort);
            }
            catch
            {
               continue;
            }
            usedPorts.Add(newPort);
            break;
         }
         return (UInt16)newPort;
      } 
      
      public int CompareTo ( Agent<T> rhs ) 
      {
         if( this.AgentID == rhs.AgentID )
            return 0 ;
         else if( this.AgentID > rhs.AgentID )
            return this.AgentID - rhs.AgentID ;
         else
            return rhs.AgentID - this.AgentID ;
      }

      public virtual void HandleDisconnection(SocketException se)
      {
         //this._active = false ;
      }

      public void Process(int iRx)
      {
         _waitAck.Set() ;
         ProcessMain(iRx) ;
      }
      
      
      public virtual void ProcessMain(int iRx)
      {
         
      }
   }
}
