﻿using System;
using System.Collections.Generic;

namespace PokerClass
{
    public class HandAnalysis
    {
        public enum HandType
        { STRAIGHTFLUSH=8, FOUROFAKIND=7, FULLHOUSE=6, FLUSH=5, STRAIGHT=4, THREEOFAKIND=3, TWOPAIR=2, PAIR=1, HIGHCARD=0 }

        public HandType handType;

        bool containsFlush;
        int flushSuit;
        public bool[] flushRanks;

        bool containsStraightFlush;
        bool containsStraight;
        public int straightHighCard;

        public int mostFrequentRank;
        public int secondMostFrequentRank;

        Card[] cards;
        int[] suitCounts;
        int[] rankCounts;
        public int[] kickers;

        public HandAnalysis(Card[] cards)
        {
            this.cards = cards;
            suitCounts = new int[Card.Suits.Length];
            rankCounts = new int[Card.Ranks.Length];
            for (int i = 0; i < cards.Length; i++)
            {
                suitCounts[(int)cards[i].suit]++;
                rankCounts[(int)cards[i].rank]++;
            }
            testContainsFlush();
            if (containsFlush)
            {
                testFlushRanks();
                testContainsStraightFlush();
                if (containsStraightFlush)
                {
                    handType = HandType.STRAIGHTFLUSH;
                    testKickers();
                    return;
                }
            }
            mostFrequentRank = -1;
            secondMostFrequentRank = -1;
            for (int i = 0; i < rankCounts.Length; i++)
            {
                if ((mostFrequentRank == -1) || rankCounts[i] >= rankCounts[mostFrequentRank])
                {
                    secondMostFrequentRank = mostFrequentRank;
                    mostFrequentRank = i;
                }
                else if ((secondMostFrequentRank == -1) || rankCounts[i] >= rankCounts[secondMostFrequentRank])
                {
                    secondMostFrequentRank = i;
                }
            }

            testContainsStraight();

            if (rankCounts[mostFrequentRank] == 4)
            {
                handType = HandType.FOUROFAKIND;
            }
            else if (rankCounts[mostFrequentRank] == 3 && rankCounts[secondMostFrequentRank] >= 2)
            {
                handType = HandType.FULLHOUSE;
            }
            else if (containsFlush)
            {
                handType = HandType.FLUSH;
            }
            else if (containsStraight)
            {
                handType = HandType.STRAIGHT;
            }
            else if (rankCounts[mostFrequentRank] == 3)
            {
                handType = HandType.THREEOFAKIND;
            }
            else if (rankCounts[mostFrequentRank] == 2)
            {
                if (rankCounts[secondMostFrequentRank] == 2)
                {
                    handType = HandType.TWOPAIR;
                }
                else
                {
                    handType = HandType.PAIR;
                }
            }
            else
            {
                handType = HandType.HIGHCARD;
            }
            testKickers();
        }

        public void testContainsFlush()
        {
            for (int i = 0; i < suitCounts.Length; i++)
            {
                if (suitCounts[i] >= 5)
                {
                    containsFlush = true;
                    flushSuit = i;
                    return;
                }
            }
            containsFlush = false;
        }

        public void testContainsStraight()
        {
            int runningCount = 0;
            for (int i = rankCounts.Length - 1; i >= 0; i--)
            {
                if (rankCounts[i] > 0)
                {
                    runningCount++;
                    if (runningCount == 5)
                    {
                        straightHighCard = i + 5;
                        containsStraight = true;
                        return;
                    }
                }
                else
                {
                    runningCount = 0;
                }
            }
            if (runningCount == 4 && rankCounts[12] > 0)
            {
                straightHighCard = 3;
                containsStraight = true;
            }
            else
            {
                containsStraight = false;
            }
        }

        public void testFlushRanks()
        {
            flushRanks = new bool[rankCounts.Length];
            for (int i = 0; i < flushRanks.Length; i++)
            {
                flushRanks[i] = false;
            }
            for (int i = 0; i < cards.Length; i++)
            {
                if ((int)cards[i].suit == flushSuit)
                {
                    flushRanks[(int)cards[i].rank] = true;
                }
            }
        }

        public void testContainsStraightFlush()
        {
            int runningCount = 0;
            for (int i = flushRanks.Length - 1; i >= 0; i--)
            {
                if (flushRanks[i])
                {
                    runningCount++;
                    if (runningCount == 5)
                    {
                        straightHighCard = i + 5;
                        containsStraightFlush = true;
                        return;
                    }
                }
                else
                {
                    runningCount = 0;
                }
            }
            if (runningCount == 4 && flushRanks[12])
            {
                straightHighCard = 3;
                containsStraightFlush = true;
            }
            else
            {
                containsStraightFlush = false;
            }
        }

        public void testKickers()
        {
            switch (handType)
            {
                case HandType.STRAIGHTFLUSH:
                case HandType.FULLHOUSE:
                case HandType.FLUSH:
                case HandType.STRAIGHT:
                    kickers = new int[0];
                    return;
                case HandType.TWOPAIR:
                    kickers = new int[1];
                    for (int i = rankCounts.Length - 1; i >= 0; i--)
                    {
                        if (i != mostFrequentRank && i != secondMostFrequentRank && rankCounts[i] > 0)
                        {
                            kickers[0] = i;
                            return;
                        }
                    }
                    throw new Exception("Reached a strange place in HandAnalysis.");
                default:
                    //case HandType.HIGHCARD:
                    //case HandType.PAIR:
                    //case HandType.THREEOFAKIND:
                    //case HandType.FOUROFAKIND:
                    int currentIndex = 0;
                    kickers = new int[5 - rankCounts[mostFrequentRank]];
                    for (int i = rankCounts.Length - 1; i >= 0; i--)
                    {
                        if (i != mostFrequentRank && rankCounts[i] > 0)
                        {
                            kickers[currentIndex++] = i;
                            if (currentIndex == kickers.Length)
                            {
                                return;
                            }
                        }
                    }
                    throw new Exception("Reached a strange place in HandAnalysis(2).");

            }
        }
    }
}