﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Net;
using System.Net.Sockets;

namespace MyTcp
{
   public interface IShowable
   {
      void ShowMessage(string s);
   }

   public class MyTcpServer<T, AgentType> 
      where T : IShowable 
      where AgentType : Agent<T> , new()
   {
      private Socket mainSocket;
      public AsyncCallback pfnWorkerCallBack;
      private Thread worker ;

      public List<Agent<T>> agentList = new List<Agent<T>>();
      
      private int port = 0 ;
      private IPAddress localAddr ;
      private T _refT ;
      
      private int _numAgent = 0 ;
      
      public MyTcpServer(int p, IPAddress ip, T r)
      {
         port = p ;
         localAddr = ip ;
         worker = new Thread(WaitForConnection) ;
         _refT = r ;
      }
      
      public void StartServer() 
      {
         worker.Start() ;
      }           
      
      public void WaitForConnection()
      {
         try
         {
            // Create the listening socket
            mainSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            IPEndPoint ipLocal = new IPEndPoint(localAddr, port);
            // Bind to local IP Address...
            mainSocket.Bind(ipLocal);
            // Start listening...
            mainSocket.Listen(100);
            // Create the call back for any client connections...
            mainSocket.BeginAccept(new AsyncCallback(OnClientConnect), null);
         }
         catch (SocketException se)
         {
            //HandleDisconnection(se);
         }
      }

      public void OnClientConnect(IAsyncResult asyn)
      {
         // Clean inactive client
         CleanInactive();
         try
         {
            // Here we complete/end the BeginAccept() asynchronous call    
            // by calling EndAccept() - which returns the reference to
            // a new Socket object
            Agent<T> incoming = new AgentType() ;
            incoming.InitAgent( _refT, ref agentList) ;
            //incoming._serverRef = this ;
            
            incoming.AgentSocket = mainSocket.EndAccept(asyn);
            // Let the worker Socket do the further processing for the 
            // just connected client
            // WaitForData(incoming.AgentSocket);
            WaitForData(incoming);         
            
            ++_numAgent;
            // Label the agent with number
            incoming.AgentID = _numAgent;
            // Add the agent with socket into agent list
            //agentList.Add(incoming);
            agentList.Add(incoming);                                   

            // Since the main Socket is now free, it can go back and wait for
            // other clients who are attempting to connect
            mainSocket.BeginAccept(new AsyncCallback(OnClientConnect), null);
         }
         catch (ObjectDisposedException)
         {
            System.Diagnostics.Debugger.Log(0, "1", "\n OnClientConnection: Socket has been closed\n");
         }
         catch (SocketException se)
         {
            //incoming.HandleDisconnection(se);
         }

      }

      // Start waiting for data from the client
      //public void WaitForData(System.Net.Sockets.Socket soc)
      public void WaitForData(Agent<T> agent)
      {
         try
         {
            if (pfnWorkerCallBack == null)
            {
               // Specify the call back function which is to be 
               // invoked when there is any write activity by the 
               // connected client
               pfnWorkerCallBack = new AsyncCallback(OnDataReceived);
            } 

            // Start receiving any data written by the connected client asynchronously
            if (agent._active)
               agent.AgentSocket.BeginReceive(agent._dataBuffer, 0, 1024, SocketFlags.None, pfnWorkerCallBack, agent);
         }
         catch (SocketException se)
         {
            agent._active = false;
            agent.HandleDisconnection(se);
         }

      }

      // This the call back function which will be invoked when the socket
      // detects any client writing of data on the stream
      public void OnDataReceived(IAsyncResult asyn)
      {
         Agent<T> agent = (Agent<T>)asyn.AsyncState;
         try
         {
            // Get the length of the bytes received
            // Complete the BeginReceive() asynchronous call by EndReceive() method
            // which will return the number of characters written to the stream by the client 
            int iRx = 0;
            if (agent._active)
               iRx = agent.AgentSocket.EndReceive(asyn);

            // Process the request according to the data received           
            agent.Process(iRx);
            if (agent._active)
            {              
               // Continue the waiting for data on the Socket
               WaitForData(agent);
            }
            else
            {
               agent = null;
               //CleanInactive();
            }
         }
         catch (ObjectDisposedException)
         {
            System.Diagnostics.Debugger.Log(0, "1", "\nOnDataReceived: Socket has been closed\n");
         }
         catch (SocketException se)
         {
            agent._active = false ;
            agent.HandleDisconnection(se);
         }
      }

      private void CleanInactive()
      {
         // clean the inactive agents in agentList
         int i = 0;
         while (i < agentList.Count)
            if (agentList[i]._active == false)
               agentList.RemoveAt(i);
            else
               ++i;
      }        
   
   }
}
