﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.Sockets;
using MyTcp;
using PokerClass;

namespace Poker_Server
{
   public enum RoomStatus
   {
      Empty = 0,
      Waiting = 1,
      Playing = 2,
   }  
   
   public enum PlayerStatus
   {
      Arriving = 0,
      Host = 1,
      Standing = 2,
      Player = 3,
      Spectator = 4,
      Betting = 7,
      BetPlaced = 8,
   } 
   
   public enum PlayerAction
   {
      Betting = 0 ,
      BetPlaced = 1 ,
   }        
   
   public class PokerAgent<T> : Agent<T> where T : IShowable
   {
      public PokerAgent() 
      {
         _onlineList = this._listRef;
      }
      public PokerAgent(T w) : base(w) 
      {
         _onlineList = this._listRef ;
      }              
          
      public static RoomStatus _room = RoomStatus.Empty ;                     
      public PlayerStatus _status = PlayerStatus.Arriving ;
      public PlayerAction _action = PlayerAction.BetPlaced ;     
      
      public int _seat = 0 ;
      public static PokerDynamics<T> _table ;      
      public static PokerAgent<T>[] _playerArr = new PokerAgent<T>[9] ;
      public static List<Agent<T> > _onlineList ;
      
      public int _bet = 0 ; 
      public int _balance = 1000 ;  
      public static int _startPos = 8 ;        
      
      // Change this in order to have different function
      public override void ProcessMain(int iRx)
      {        
         // Clear the response buffer first
         if (_responseBuffer != null)
            Array.Clear(_responseBuffer, 0, _responseBuffer.Length);
         
         /*
         // Process the incoming data _dataBuffer
         string data = System.Text.Encoding.Unicode.GetString(_dataBuffer, 0, iRx);
         ShowMessage(data + " Agend ID: " + this.AgentID);
          */   
            //   Modify the table here. 
         switch( _room ) {
            case RoomStatus.Empty :
            case RoomStatus.Waiting :                     
               switch( _status ) {
                  case PlayerStatus.Arriving :
                     Arrival() ;
                     break ;
                  case PlayerStatus.Host :
                     HostWaiting() ;
                     break ;
                  case PlayerStatus.Spectator:
                     ChooseSeat() ;
                     break ;
                  case PlayerStatus.Player :
                     PlayerWaiting() ;
                     break ;
               }
               break ;
            case RoomStatus.Playing :
               switch( _status ) {
                  case PlayerStatus.Arriving :
                     Arrival() ;
                     break ;
                  case PlayerStatus.Spectator :
                     ChooseSeat() ;      
                     _table.UpdateNewcomer(_seat) ;                 
                     break ;
               }
               switch( _action ) {
                  case PlayerAction.Betting :
                     Bet() ;
                     break ;                     
               }
               break ;
            default :
               break ;
         }
         
         // After the incoming data is processed, clear _dataBuffer
         Array.Clear(_dataBuffer, 0, _dataBuffer.Length);
      }         
      
      public void Arrival()
      {
         if( _room == RoomStatus.Empty ) {
            _status = PlayerStatus.Host ;
            _room = RoomStatus.Waiting ;
            for( int i = 0 ; i < 9 ; ++i ) 
               _playerArr[i] = null ;
            _playerArr[0] = this ;
         }
         else if( _room == RoomStatus.Waiting ) {        
            _status = PlayerStatus.Spectator ;
            Send(EmptySeats()) ;
         }
         else if( _room == RoomStatus.Playing ) {
            _status = PlayerStatus.Spectator ;
            Send(EmptySeats()) ;
         }
      }
      
      public int EmptySeats(ref int[] seats)
      {
         int num = 0 ;         
         for( int i = 0 ; i < 9 ; ++i )
            if( _playerArr[i] == null ) {
               ++num ;               
            }
         seats = new int[num] ;
         int j = 0 ;
         for( int i = 0 ; i < 9 ; ++i )
            if( _playerArr[i] == null ) {
               seats[j] = i ;
               ++j ;
            }
         return num ;
      }
      
      public byte[] EmptySeats()
      {
         int[] seats = null ;
         int num = EmptySeats(ref seats) ;
         byte[] response = new byte[(2+num)*4] ;
         Buffer.BlockCopy(BitConverter.GetBytes((int)10), 0, response, 0, 4) ;
         Buffer.BlockCopy(BitConverter.GetBytes(num), 0, response, 4 , 4) ;
         for( int i = 0 ; i < num ; ++i )
            Buffer.BlockCopy(BitConverter.GetBytes(seats[i]), 0, response, 8+4*i, 4) ;
         return response ;
      }
      
      public void ChooseSeat()
      {
         _seat = BitConverter.ToInt32(_dataBuffer, 0) ;
         if( _playerArr[_seat] == null ) {
            _status = PlayerStatus.Player ;
            _playerArr[_seat] = this ;            
         }
         else {
            string error = "The seat is occupied" ;
            Send(System.Text.Encoding.Unicode.GetBytes(error)) ;
         }            
      }
      
      public void PlayerWaiting() {}
      
      public void HostWaiting()
      {
         byte control = _dataBuffer[0] ;
         byte[] response = new byte[1] ;
         // Check signal for starting the round
         if( control == byte.Parse("1") ) {
            if( !CheckSpectator() ) { // There are no spectator
               _room = RoomStatus.Playing ;
               _table = new PokerDynamics<T>(GetNextStartPos(), _playerArr);
            }
         }
      }
      
      public int GetNextStartPos()
      {
         while(true) {
            if( _startPos < 8 )
               _startPos++ ;
            else
               _startPos = 0 ;
            if( _playerArr[_startPos] != null )
               return _startPos ;
         }
      }
      
      private bool CheckSpectator()
      {
         foreach( PokerAgent<T> pa in _listRef )
            if( pa._status == PlayerStatus.Spectator )
               return true ;
         return false ;
      }
      
      public void Bet()
      {
         _bet = BitConverter.ToInt32(_dataBuffer, 0) ;
         _action = PlayerAction.BetPlaced ;
      }
      
      public void GetBet()
      {
         _action = PlayerAction.Betting;
         SendWait(BitConverter.GetBytes((int)21)) ;
      }

      public override void HandleDisconnection(SocketException se)
      {
         
      }
      
      public override string ToString()
      {
 	      return _seat.ToString() ;
      }
         
   }
}
