﻿using System;
using System.Collections.Generic;
using System.Threading;
using Poker_Server;
using MyTcp;

namespace PokerClass
{
   public class PokerDynamics<T> where T : IShowable
    {
        public enum TableState
        { INIT=0, PREFLOP_DEAL, PREFLOP_BET, FLOP_DEAL, FLOP_BET, TURN_DEAL, TURN_BET, RIVER_DEAL, RIVER_BET, FINAL, END }

        public TableState tableState = TableState.INIT;

        public Card[][] hole =  new Card[9][];
        Card[][] playerHands = new Card[9][];
        public Card[] board;
        public List<Card> shuffledDeck;

        /* 
         * 9 seats for maximum 9 players
         * 
         * "int" is used to represent players temporally
         * might be substituted with "agent" class instead
         * 
         * 0 or null represents an empty seat
         * 
         */
        //public int[] players = new int[9];
        public PokerAgent<T>[] players ;
        public PokerAgent<T>[] backupArr ;
        public int numberOfPlayers = 0;

        public bool awaitAct;    // true if waiting for player's bet
        public int seatToAct;    // 0~8
        public int dealerButton; // 0~8

        public static int SMALLBLIND = 10;
        public static int BIGBLIND = 20;

        public int[] inPot = new int[9]{0,0,0,0,0,0,0,0,0}; // contribution from each player of a stage of bet
        public bool[] played = new bool[9] { false, false, false, false, false, false, false, false, false };
        public int jackPot = 0; // total price money
        public int currentBet = 0;

        public PokerDynamics( int dealerButtonInput, PokerAgent<T>[] playersInput )
        {
            dealerButton = dealerButtonInput;
            backupArr = playersInput ;
            players = (PokerAgent<T>[])playersInput.Clone() ;  

            Thread pokerCore = new Thread(new ThreadStart(PokerCore));
            pokerCore.Start();
        }

        public void PokerCore()
        {
            while (tableState != TableState.END)
            {
                if (!awaitAct)
                    switch (tableState)
                    {
                        case TableState.INIT:
                            // deal card
                            dealCard();
                            break;
                        case TableState.PREFLOP_DEAL:
                            // notify hole cards
                            notifyCards();
                            break;
                        case TableState.PREFLOP_BET:
                            // force blind
                            // start betting from dealer + 3
                            takeBets();
                            break;
                        case TableState.FLOP_DEAL:
                            // notify board cards 0~2
                            notifyCards();
                            break;
                        case TableState.FLOP_BET:
                            // start betting from dealer + 1
                            takeBets();
                            break;
                        case TableState.TURN_DEAL:
                            // notify board card 3
                            notifyCards();
                            break;
                        case TableState.TURN_BET:
                            // start betting from dealer + 1
                            takeBets();
                            break;
                        case TableState.RIVER_DEAL:
                            // notify board card 4
                            notifyCards();
                            break;
                        case TableState.RIVER_BET:
                            // start betting from dealer + 1
                            takeBets();
                            break;
                        case TableState.FINAL:
                            // determine winner
                            // price money
                            // terminate round
                            determineWinner();
                            break;
                    }
            }
            
            endRound() ;
        }

        public void dealCard()
        {                      
            foreach (PokerAgent<T> player in players)
                if (player != null)
                    numberOfPlayers++;
                    
            // TODO
            //showMessage(numberOfPlayers + " players are playing the game", 9);
            UpdateBalance() ; 
            
            // TODO 
            SendTableState() ;                       
            //showMessage("Dealing Cards", 9);

            shuffledDeck = Card.getShuffledCards();

            for (int player = 0; player < 9; player++)
                if (players[player] != null)
                {
                    hole[player] = new Card[2];
                    hole[player][0] = shuffledDeck[0];
                    hole[player][1] = shuffledDeck[1];
                    shuffledDeck.RemoveRange(0, 2);
                }

            board = new Card[5];

            board[0] = shuffledDeck[0];
            board[1] = shuffledDeck[1];
            board[2] = shuffledDeck[2];
            board[3] = shuffledDeck[3];
            board[4] = shuffledDeck[4];
            shuffledDeck.RemoveRange(0, 5);

            tableState++;
        }

        public void notifyCards()
        {
            SendTableState() ;
            //showMessage("\nNotifying Cards at " + tableState, 9);

            switch (tableState)
            {
                case TableState.PREFLOP_DEAL:
                    for (int player = 0; player < 9; player++)
                        if (players[player] != null)
                        {
                            SendHole(player) ;
                            /*
                            showMessage("Cards for player " + player + ":" , player);
                            showMessage(hole[player][0].CardString(), player);
                            showMessage(hole[player][1].CardString(), player);
                            */
                        }
                    break;
                case TableState.FLOP_DEAL:
                    /*
                    showMessage("Revealing 3 board cards:", 9);
                    showMessage(board[0].CardString(),9);
                    showMessage(board[1].CardString(),9);
                    showMessage(board[2].CardString(),9);
                     */
                    BoardThree() ;
                    break;
                case TableState.TURN_DEAL:
                    /*
                    showMessage("Revealing 1 board card:",9);
                    showMessage(board[3].CardString(),9);*/
                    BoardOne(3) ;
                    break;
                case TableState.RIVER_DEAL:
                    /*
                    showMessage("Revealing 1 board card:",9);
                    showMessage(board[4].CardString(),9);
                     */
                    BoardOne(4) ;
                    break;
                default:
                    processError();
                    break;
            }


            tableState++;
        }

        public void takeBets()
        {
            //showMessage("\nTaking Bets at " + tableState,9);
            SendTableState() ;

            switch (tableState)
            {
                case TableState.PREFLOP_BET:

                    //showMessage("Start betting...",9);

                    seatToAct = (dealerButton + 1) % 9;
                    while(players[seatToAct] == null)
                        seatToAct = (seatToAct + 1) % 9;
                    inPot[seatToAct] = SMALLBLIND;
                    jackPot += SMALLBLIND;
                    // Collect bet
                    backupArr[seatToAct]._balance -= SMALLBLIND ;                    
                    
                    // Show player's inpot to others
                    SendBet(seatToAct, inPot[seatToAct], backupArr[seatToAct]._balance) ;                    
                    //showMessage("Bet from player " + players[seatToAct] + ": " + inPot[seatToAct],9);

                    seatToAct = (seatToAct + 1) % 9;
                    while (players[seatToAct] == null)
                        seatToAct = (seatToAct + 1) % 9;
                    inPot[seatToAct] = BIGBLIND;
                    jackPot += BIGBLIND;
                    // Collect bet
                    backupArr[seatToAct]._balance -= BIGBLIND ;
                    
                    // Show player's inpot to others
                    SendBet(seatToAct, inPot[seatToAct], backupArr[seatToAct]._balance); 
                    //showMessage("Bet from player " + players[seatToAct] + ": " + inPot[seatToAct],9);

                    currentBet = BIGBLIND;
                    seatToAct = (seatToAct + 1) % 9;

                    takeABet();

                    break;
                case TableState.FLOP_BET:

                    //showMessage("Start betting...",9);
                    
                    seatToAct = (dealerButton + 1) % 9;
                    while (players[seatToAct] == null)
                        seatToAct = (seatToAct + 1) % 9;

                    currentBet = 0;
                    takeABet();

                    break;
                case TableState.TURN_BET:

                    //showMessage("Start betting...",9);
                    
                    seatToAct = (dealerButton + 1) % 9;
                    while (players[seatToAct] == null)
                        seatToAct = (seatToAct + 1) % 9;

                    currentBet = 0;
                    takeABet();

                    break;
                case TableState.RIVER_BET:

                    //showMessage("Start betting...",9);
                    
                    seatToAct = (dealerButton + 1) % 9;
                    while (players[seatToAct] == null)
                        seatToAct = (seatToAct + 1) % 9;

                    currentBet = 0;
                    takeABet();

                    break;
                default:
                    processError();
                    break;
            }

            if (tableState != TableState.END)
            {
                tableState++;
                
                SendJackpot() ;
                //showMessage("Total JackPot : " + jackPot,9);
            }
        }

        public void takeABet()
        {
            int tmp;

            while (tableState != TableState.END)
            {
                if (players[seatToAct] != null)
                {
                    if (inPot[seatToAct] == currentBet && played[seatToAct] == true)
                        break;

                    //take bets
                    //Console.Write("Bet from player " + players[seatToAct] + ": ");
                    
                    // Change players[seatToAct] status to Betting
                    players[seatToAct].GetBet() ;
                    //showMessage("Bet from player " + players[seatToAct] + ": ",9) ;
                                        
                    // Suspend the table and wait for a player to place a bet
                    while(true)
                    {
                       if( players[seatToAct]._action == PlayerAction.BetPlaced ) {
                          tmp = players[seatToAct]._bet ;                                                    
                          //showMessage("Bet from player " + players[seatToAct] + ": " + tmp ,9);                          
                          break ;
                       }
                    }
                    //tmp = int.Parse(Console.ReadLine());

                    if (tmp == 0 && currentBet != 0)
                    {
                        // Show player's current inpot
                        SendBet(seatToAct, tmp, backupArr[seatToAct]._balance);
                        players[seatToAct] = null;
                        checkPrematureWinner();
                    }
                    else if (tmp < currentBet)
                        continue;
                    else
                    {
                        jackPot += tmp - inPot[seatToAct];
                        // Collect bet
                        backupArr[seatToAct]._balance -= (tmp - inPot[seatToAct]) ;
                        // Show player's current inpot
                        SendBet(seatToAct, tmp, backupArr[seatToAct]._balance);
                                                                                            
                        currentBet = tmp;
                        inPot[seatToAct] = tmp;
                        played[seatToAct] = true;
                    }
                    
                }
                seatToAct = (seatToAct + 1) % 9;
            }

            for (int i = 0; i < 9; i++)
            {
                inPot[i] = 0;
                played[i] = false;
            }
        }

        public void checkPrematureWinner()
        {
            int playerCount = 0;

            foreach (PokerAgent<T> player in players)
                if (player != null)
                    playerCount++;

            if (playerCount == 1)
                foreach (PokerAgent<T> player in players)
                    if (player != null)
                        givePrice(player);
        }

        public void determineWinner()
        {
           //showMessage("\nDetermining Winner",9);
           SendTableState();

           for (int i = 0; i < 9; i++)
              if (players[i] != null)
              {
                 playerHands[i] = new Card[7];
                 playerHands[i][0] = hole[i][0];
                 playerHands[i][1] = hole[i][1];

                 for (int j = 0; j < 5; j++)
                    playerHands[i][j + 2] = board[j];
              }

           for (int i = 0; i < 9; i++)
              for (int j = i + 1; j < 9; j++)
                 if (players[i] != null && players[j] != null)
                 {
                    switch (determineWinner(playerHands[i], playerHands[j]))
                    {
                       case 0:
                          players[j] = null;
                          break;
                       case 1:
                          players[i] = null;
                          break;
                       default: //tie
                          break;
                    }
                 }

           givePrices();
        }

        public static int determineWinner(Card[] hand1, Card[] hand2)
        {

           HandAnalysis seat0 = new HandAnalysis(hand1);
           HandAnalysis seat1 = new HandAnalysis(hand2);

           if ((int)seat0.handType > (int)seat1.handType)
           {
              return 0;
           }
           if ((int)seat1.handType < (int)seat0.handType)
           {
              return 1;
           }
           switch (seat0.handType)
           {
              case HandAnalysis.HandType.STRAIGHTFLUSH:
              case HandAnalysis.HandType.STRAIGHT:
                 if (seat0.straightHighCard > seat1.straightHighCard)
                 {
                    return 0;
                 }
                 if (seat1.straightHighCard > seat0.straightHighCard)
                 {
                    return 1;
                 }
                 return -1;
              case HandAnalysis.HandType.FLUSH:
                 int numCards = 0;
                 for (int i = 12; i >= 0; i--)
                 {
                    if (seat0.flushRanks[i])
                    {
                       if (seat1.flushRanks[i])
                       {
                          numCards++;
                          if (numCards == 5)
                          {
                             return -1;
                          }
                       }
                       else
                       {
                          return 0;
                       }
                    }
                    else if (seat1.flushRanks[i])
                    {
                       return 1;
                    }
                 }
                 return -1;
              case HandAnalysis.HandType.FOUROFAKIND:
              case HandAnalysis.HandType.THREEOFAKIND:
              case HandAnalysis.HandType.PAIR:
              case HandAnalysis.HandType.HIGHCARD:
                 if (seat0.mostFrequentRank > seat1.mostFrequentRank)
                 {
                    return 0;
                 }
                 if (seat0.mostFrequentRank < seat1.mostFrequentRank)
                 {
                    return 1;
                 }
                 return kickerTest(seat0, seat1);
              case HandAnalysis.HandType.FULLHOUSE:
              case HandAnalysis.HandType.TWOPAIR:
                 if (seat0.mostFrequentRank > seat1.mostFrequentRank)
                 {
                    return 0;
                 }
                 if (seat0.mostFrequentRank < seat1.mostFrequentRank)
                 {
                    return 1;
                 }
                 if (seat0.secondMostFrequentRank > seat1.secondMostFrequentRank)
                 {
                    return 0;
                 }
                 if (seat0.secondMostFrequentRank < seat1.secondMostFrequentRank)
                 {
                    return 1;
                 }
                 return kickerTest(seat0, seat1);
              default:
                 return -1;
           }
        }

        public static int kickerTest(HandAnalysis seat0, HandAnalysis seat1)
        {
           for (int i = 0; i < seat0.kickers.Length; i++)
           {
              if (seat0.kickers[i] > seat1.kickers[i])
              {
                 return 0;
              }
              if (seat0.kickers[i] < seat1.kickers[i])
              {
                 return 1;
              }
           }
           return -1;
        }    

        public void givePrice(PokerAgent<T> winner)
        {
            //showMessage("Player " + winner._seat + " won the hand",9);                        
            SendWinner(winner._seat) ;
            //showMessage("Jackpot: " + jackPot,9);
            SendJackpot() ;
            
            // Give money
            backupArr[winner._seat]._balance += jackPot ;
            
            jackPot = 0;
            tableState = TableState.END;
        }

        public void givePrices()
        {
           int playerCount = 0;

           foreach (PokerAgent<T> player in players)
              if (player != null)
                 playerCount++;
                 
           // Reveal winners' holes
           int[] winnerArr = new int[playerCount] ;
           int index = 0 ;
           foreach(PokerAgent<T> player in players)
              if (player != null) {
                 winnerArr[index] = player._seat ;
                 index++ ;
              }                          

           //showMessage(playerCount + " player(s) share the jackpot", 9);
           //showMessage("Jackpot: " + jackPot, 9);
           

           HandAnalysis playerHand = null;
           for (int i = 0; i < 9; i++)
              if (players[i] != null)
              {
                 playerHand = new HandAnalysis(playerHands[i]);
                 // TODO
                 //showMessage("Player " + players[i] + " won " + jackPot / playerCount + " with " + playerHand.handType, 9);
                 backupArr[i]._balance += jackPot / playerCount ;
              }

           WinnerHole(winnerArr, playerCount, playerHand.handType);
           SendJackpot();
           
           jackPot = 0;
           tableState = TableState.END;
        }
        
        public void endRound()
        {
            UpdateBalance() ;
            //showMessage("Round finished", 9) ;
            SendTableState() ;
            
            Poker_Server.PokerAgent<T>._room = RoomStatus.Waiting ;
        }
        
        public PokerAgent<T> getAliveAgent()
        {
            foreach( PokerAgent<T> pa in players )
               if( pa != null )
                  return pa ;
            return null ;
        }

        public void processError()
        {
            showMessage("An error has occured at " + tableState,9);
        }

        public void showMessage(string message, int seat)
        {
            //Console.WriteLine(message);
            // Write message to server window
            // Pass message to all the players
            
            backupArr[0]._windowRef.ShowMessage(message) ;
            
            /*
            if( seat == 9 ) {
               PokerAgent<T> alive = getAliveAgent() ;
               //foreach( PokerAgent<T> pa in PokerAgent<T>._onlineList ) 
               foreach( PokerAgent<T> pa in alive._listRef )
               //foreach( PokerAgent<T> pa in players )
                  if( pa != null )
                     pa.Send(System.Text.Encoding.Unicode.GetBytes(message+"\n") ) ;
            }
            else
               players[seat].Send( System.Text.Encoding.Unicode.GetBytes(message+"\n") ) ;
            */
        }
        
        public void SendAll(byte[] message) 
        {
           PokerAgent<T> alive = getAliveAgent() ;              
           foreach( PokerAgent<T> pa in backupArr )                  
               if( pa != null )
                     pa.SendWait(message) ;
        }
        
        public void SendWinner(int seat)
        {
           byte[] message = new byte[8] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)28), 0, message, 0, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(seat), 0, message, 4, 4) ;
           SendAll(message) ;
        }
        
        public void BoardThree()
        {
           byte[] message = new byte[28] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)22), 0, message, 0, 4) ;
           for( int i = 0 ; i < 3 ; ++i ) {
              Buffer.BlockCopy(BitConverter.GetBytes(board[i].SuitInt()), 0, message, 4+8*i, 4) ;
              Buffer.BlockCopy(BitConverter.GetBytes(board[i].RankInt()), 0, message, 8+8*i, 4) ;
           }
           SendAll(message) ;
        }
        
        public void BoardOne(int card)
        {
           byte[] message = new byte[12] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)24), 0, message, 0, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(board[card].SuitInt()), 0, message, 4, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(board[card].RankInt()), 0, message, 8, 4) ;
           SendAll(message) ;
        }
        
        public void SendHole(int seat)
        {
           byte[] message = new byte[20] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)23), 0, message, 0, 4) ;
           for( int i = 0 ; i < 2 ; ++i ) {
              Buffer.BlockCopy(BitConverter.GetBytes(hole[seat][i].SuitInt()), 0, message, 4+8*i, 4) ;
              Buffer.BlockCopy(BitConverter.GetBytes(hole[seat][i].RankInt()), 0, message, 8+8*i, 4) ;
           }
           players[seat].SendWait(message) ;           
        }
        
        public void WinnerHole(int[] winners, int num, HandAnalysis.HandType htype) 
        {
           byte[] message = new byte[12+20*num] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)25), 0, message, 0, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(num), 0, message, 4, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)htype), 0, message, 8, 4) ;
           for( int i = 0 ; i < num ; ++i ) {
              for( int j = 0 ; j < 2 ; ++j ) {
                 Buffer.BlockCopy(BitConverter.GetBytes(winners[i]), 0, message, 12+20*i, 4) ;
                 Buffer.BlockCopy(BitConverter.GetBytes(hole[winners[i]][j].SuitInt()), 
                                  0, message, 16+20*i+8*j, 4) ;
                 Buffer.BlockCopy(BitConverter.GetBytes(hole[winners[i]][j].RankInt()), 
                                  0, message, 20+20*i+8*j, 4) ;
              }
           }
           SendAll(message) ;
        }
        
        public void SendBet(int seat, int pot, int currentMoney)
        {
           byte[] message = new byte[16] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)26), 0, message, 0, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(seat), 0, message, 4, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(pot), 0, message, 8, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(currentMoney), 0, message, 12, 4) ;
           SendAll(message) ;
        }
        
        public void SendJackpot()
        {
           byte[] message = new byte[8] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)27), 0, message, 0, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(jackPot), 0, message, 4, 4) ;
           SendAll(message) ;
        }
        
        public void SendTableState()
        {  
           byte[] message = new byte[8] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)11), 0, message, 0, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)tableState), 0, message, 4, 4) ;
           SendAll(message) ;
        }
        
        public void UpdateBalance()
        {
           byte[] message = new byte[8+8*numberOfPlayers] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)12), 0, message, 0, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(numberOfPlayers), 0, message, 4, 4) ;
           int i = 0 ;
           foreach(PokerAgent<T> pa in backupArr)
              if( pa != null ) {
                 Buffer.BlockCopy(BitConverter.GetBytes(pa._seat), 0, message, 8+8*i, 4) ;
                 Buffer.BlockCopy(BitConverter.GetBytes(pa._balance), 0, message, 12+8*i, 4) ;
                 ++i ;
              }
           SendAll(message) ;
        }
        
        public void UpdateNewcomer(int seat)
        {
           int numCards = 0 ;
           switch( tableState ) {
              case TableState.INIT:
              case TableState.PREFLOP_DEAL:
              case TableState.PREFLOP_BET:
                 numCards = 0 ;
                 break ;
              case TableState.FLOP_DEAL:
              case TableState.FLOP_BET:
                 numCards = 3 ;
                 break ;
              case TableState.TURN_DEAL:
              case TableState.TURN_BET:
                 numCards = 4 ;
                 break ;
              case TableState.RIVER_DEAL:
              case TableState.RIVER_BET:
              case TableState.FINAL:
              default :
                 numCards = 5 ;
                 break ;
           }
           
           byte[] message = new byte[8+8*numberOfPlayers+4+8*numCards] ;
           Buffer.BlockCopy(BitConverter.GetBytes((int)13), 0, message, 0, 4) ;
           Buffer.BlockCopy(BitConverter.GetBytes(numberOfPlayers), 0, message, 4, 4);
           int i = 0 ;
           foreach(PokerAgent<T> pa in backupArr)
              if( pa != null ) {
                 Buffer.BlockCopy(BitConverter.GetBytes(pa._seat), 0, message, 8+8*i, 4) ;
                 Buffer.BlockCopy(BitConverter.GetBytes(pa._balance), 0, message, 12+8*i, 4) ;
                 ++i ;
              }
           Buffer.BlockCopy(BitConverter.GetBytes(numCards), 0, message, 8 + 8 * numberOfPlayers, 4);
           for( int j = 0 ; j < numCards ; ++j ) {
              Buffer.BlockCopy(BitConverter.GetBytes(board[j].SuitInt()), 0
                              , message, 12+8*numberOfPlayers+8*j, 4) ;
              Buffer.BlockCopy(BitConverter.GetBytes(board[j].RankInt()), 0
                              , message, 16+8*numberOfPlayers+8*j, 4) ;
           }
           backupArr[seat].SendWait(message) ;
        }
              
           
           
    }
    
}